"use strict";
const path_1 = require("path");
const fs_1 = require("fs");
const http_1 = require("http");
const command_1 = require("@oclif/command");
const plantumlEncoder = require("plantuml-encoder");
const typeorm_1 = require("typeorm");
const parseFlagsArray = (items) => (items || '')
    .split(',')
    .map(item => item.trim())
    .filter((item) => item.length)
    .map(item => {
    const matches = item.match(/^\/(.+)\/$/);
    if (matches) {
        return new RegExp(matches[1]);
    }
    return item;
});
const isInArray = (target, items) => {
    for (const item of items) {
        if (typeof item === 'string') {
            if (item === target) {
                return true;
            }
        }
        else if (item instanceof RegExp) {
            if (target.match(item)) {
                return true;
            }
        }
    }
    return false;
};
var NamesEnum;
(function (NamesEnum) {
    NamesEnum["Entities"] = "entities";
    NamesEnum["Tables"] = "tables";
})(NamesEnum || (NamesEnum = {}));
class TypeormUmlCommand extends command_1.Command {
    constructor() {
        super(...arguments);
        this.outEntities = false;
        this.outTables = false;
    }
    async run() {
        try {
            const { args, flags } = this.parse(TypeormUmlCommand);
            this.flags = flags;
            this.exclude = parseFlagsArray(this.flags.exclude);
            this.include = parseFlagsArray(this.flags.include);
            this.outEntities = this.flags.names.includes(NamesEnum.Entities);
            this.outTables = this.flags.names.includes(NamesEnum.Tables);
            if (flags.uml) {
                const uml = await this.getUml(args.configName);
                process.stdout.write(`${uml}\n`);
            }
            else {
                const url = await this.getUrl(args.configName);
                if (flags.download) {
                    await this.download(url, flags.download);
                }
                else {
                    process.stdout.write(`${url}\n`);
                }
            }
        }
        catch (e) {
            this.error(e.message);
        }
    }
    async getUml(configName) {
        const connectionOptionsReader = new typeorm_1.ConnectionOptionsReader({
            root: process.cwd(),
            configName,
        });
        const connectionOptions = await connectionOptionsReader.get(this.flags.connection);
        const connection = await typeorm_1.createConnection(connectionOptions);
        const uml = this.buildUml(connection);
        connection.close();
        return uml;
    }
    async getUrl(configName) {
        const uml = await this.getUml(configName);
        const encodedUml = plantumlEncoder.encode(uml);
        const format = encodeURIComponent(this.flags.format);
        const schema = encodeURIComponent(encodedUml);
        return `http://www.plantuml.com/plantuml/${format}/${schema}`;
    }
    download(url, filename) {
        const path = !path_1.isAbsolute(filename) ? path_1.resolve(process.cwd(), filename) : filename;
        return new Promise((resolve) => {
            http_1.get(url, (response) => {
                response.pipe(fs_1.createWriteStream(path));
                response.on('end', resolve);
            });
        });
    }
    buildUml(connection) {
        let uml = `@startuml\n\n`;
        if (this.outEntities && this.outTables) {
            uml += `!define table(ename, tname) entity "<b>ename</b>\\n<font size="10" color="gray">tname</font>" as ename\n`;
            uml += `!define column(pname, tname, type) pname    <font color="#a9a9a9">tname</font><font color="gray">: type</font>\n`;
        }
        else if (this.outEntities) {
            uml += `!define table(ename, tname) entity ename\n`;
            uml += `!define column(pname, tname, type) pname<font color="gray">: type</font>\n`;
        }
        else if (this.outTables) {
            uml += `!define table(ename, tname) entity "tname" as ename\n`;
            uml += `!define column(pname, tname, type) tname<font color="gray">: type</font>\n`;
        }
        else {
            throw new Error('Please specify what names to output using --names argument');
        }
        uml += `hide stereotypes\n`;
        uml += `hide methods\n`;
        uml += `hide circle\n\n`;
        if (this.flags.monochrome) {
            uml += `skinparam monochrome true\n\n`;
        }
        let foreignUmls = "";
        for (let i = 0, len = connection.entityMetadatas.length; i < len; i++) {
            const entity = connection.entityMetadatas[i];
            if (!this.isEntityIncluded(entity)) {
                continue;
            }
            const [classUml, foreignUml] = this.buildClass(entity, connection);
            uml += classUml;
            foreignUmls += foreignUml;
        }
        uml += foreignUmls;
        uml += `@enduml\n`;
        return uml;
    }
    getEntityKey(entity) {
        return entity.name;
    }
    getColumnKey(column) {
        if (this.outEntities) {
            return column.propertyName;
        }
        else {
            return column.databaseName;
        }
    }
    isEntityIncluded(entity) {
        if (isInArray(entity.name, this.exclude)) {
            return false;
        }
        if (this.include.length > 0 && !isInArray(entity.name, this.include)) {
            return false;
        }
        return true;
    }
    buildClass(entity, connection) {
        let uml = `\ntable(${this.getEntityKey(entity)}, ${entity.tableNameWithoutPrefix}) {\n`;
        for (let i = 0, len = entity.columns.length; i < len; i++) {
            uml += this.buildColumn(entity.columns[i], entity, connection);
        }
        uml += `}\n\n`;
        let foreignKeysUml = "";
        for (let i = 0, len = entity.foreignKeys.length; i < len; i++) {
            foreignKeysUml += this.buildForeignKeys(entity.foreignKeys[i], entity);
        }
        return [uml, foreignKeysUml];
    }
    buildColumn(column, entity, connection) {
        let prefix = '';
        if (column.isPrimary) {
            prefix = '+';
        }
        else if (Array.isArray(entity.indices) && entity.indices.length > 0) {
            const index = entity.indices.find((idx) => idx.columns.map(column => column.databaseName).includes(column.databaseName));
            if (index) {
                prefix = index.isUnique ? '~' : '#';
            }
        }
        let length = this.getColumnLength(column);
        const type = connection.driver.normalizeType(column);
        if (!length && connection.driver.dataTypeDefaults[type]) {
            length = this.getColumnLength(connection.driver.dataTypeDefaults[type]);
        }
        if (length) {
            length = `(${length})`;
        }
        const columnType = `${type.toUpperCase()}${length}`;
        const columnInfo = `column(${column.propertyName}, ${column.databaseName}, ${columnType})`;
        return `\t${prefix}${columnInfo}\n`;
    }
    buildForeignKeys(foreignKey, entity) {
        if (!this.isEntityIncluded(foreignKey.referencedEntityMetadata)) {
            return '';
        }
        const referencedEntityKey = this.getEntityKey(foreignKey.referencedEntityMetadata);
        if (foreignKey.columns.length === 1) {
            const column = foreignKey.columns[0];
            return `${this.getEntityKey(entity)}::${this.getColumnKey(column)} --> ${referencedEntityKey}\n\n`;
        }
        else {
            return `${this.getEntityKey(entity)} --> ${referencedEntityKey}\n\n`;
        }
    }
    getColumnLength(column) {
        if (column.length) {
            return column.length;
        }
        if (column.width) {
            return column.width.toString();
        }
        if (column.precision) {
            if (column.scale) {
                return `${column.precision}, ${column.scale}`;
            }
            return column.precision.toString();
        }
        return '';
    }
}
TypeormUmlCommand.description = 'Generates a database UML diagram based on Typeorm entities.';
TypeormUmlCommand.args = [
    {
        name: 'configName',
        required: false,
        description: 'Path to the Typeorm config file.',
        default: 'ormconfig.json',
    },
];
TypeormUmlCommand.flags = {
    connection: command_1.flags.string({
        char: 'c',
        description: 'The connection name.',
        default: 'default',
    }),
    format: command_1.flags.string({
        char: 'f',
        description: 'The diagram file format.',
        default: 'png',
        options: ['png', 'svg', 'txt'],
    }),
    monochrome: command_1.flags.boolean({
        description: 'Whether or not to use monochrome colors.',
        default: false,
    }),
    download: command_1.flags.string({
        char: 'd',
        description: 'The filename where to download the diagram.',
    }),
    uml: command_1.flags.boolean({
        char: 'u',
        description: 'Outputs plantuml syntax instead of the url.',
        default: false
    }),
    names: command_1.flags.string({
        char: 'n',
        multiple: true,
        options: [NamesEnum.Entities, NamesEnum.Tables],
        description: 'Specify which names to output: typeorm names, database names or both.',
        default: [NamesEnum.Tables]
    }),
    exclude: command_1.flags.string({
        char: 'e',
        description: 'Comma-separated list of entities to exclude from the diagram.',
    }),
    include: command_1.flags.string({
        char: 'i',
        description: 'Comma-separated list of entities to include into the diagram.',
    }),
};
module.exports = TypeormUmlCommand;
